// Copyright 2018 PDFium Authors. All rights reserved.
// Use of this source code is governed by a BSD-style license that can be
// found in the LICENSE file.

#include "core/fpdfapi/page/cpdf_transparency.h"

CPDF_Transparency::CPDF_Transparency() = default;

CPDF_Transparency::CPDF_Transparency(const CPDF_Transparency& other) = default;
