// Copyright 2019 PDFium Authors. All rights reserved.
// Use of this source code is governed by a BSD-style license that can be
// found in the LICENSE file.
//
// Original code copyright 2014 Foxit Software Inc. http://www.foxitsoftware.com

#include "core/fxcodec/jbig2/JBig2_DocumentContext.h"

#include "core/fxcodec/jbig2/JBig2_Image.h"
#include "core/fxcodec/jbig2/JBig2_SymbolDict.h"

JBig2_DocumentContext::JBig2_DocumentContext() = default;

JBig2_DocumentContext::~JBig2_DocumentContext() = default;
